use serde::{Serialize, Deserialize};
use crate::defs::{
  Cluster,
  LoginUser,
  SSHAccess,
	Cntrllr,
  NetworkInterface,
  StorageDevice,
};
use crate::clouds::defs::{TskSrvc};

#[derive(Eq, PartialEq, Copy, Clone, Debug, Serialize,  Deserialize)]
#[allow(non_camel_case_types)]
pub enum ServerPlan {
  ubuntu_1xCPU_1GB,
  manual,
}

impl Default for ServerPlan {
 fn default() -> Self {
    ServerPlan::manual
 }
}
impl ServerPlan {
 pub fn get_plan (&self) -> String {
   match self {
    ServerPlan::ubuntu_1xCPU_1GB => String::from("1xCPU-1GB"),
    ServerPlan::manual => String::from("manual"),
   }
 }
}

#[allow(non_camel_case_types)]
#[derive(Eq, PartialEq, Copy, Clone,Debug, Serialize,  Deserialize)]
pub enum Zone {
  local,
}

impl Default for Zone {
  fn default() -> Self {
     Zone::local
  }
} 
impl Zone {
 pub fn get_zone (&self) -> String {
   match self {
     Zone::local => String::from("local"),
   }
 }
}

#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize)]
pub struct ServerInstance {
  #[serde(default)]
  pub cluster: Cluster,
  pub hostname: String,
  pub title: String,
  #[serde(default)]
  pub loginUser: LoginUser,
  #[serde(default)]
  pub metadata: bool,
  #[serde(default)]
  pub plan: ServerPlan,
  #[serde(default)]
  pub zone: Zone,
  pub arch: String,
  #[serde(default)]
  pub sshAccess: SSHAccess,
  #[serde(default)]
  pub networks: Vec<NetworkInterface>,
  #[serde(default)]
  pub storageDevices: Vec<StorageDevice>,
  #[serde(default)]
  pub timeZone: String,
  #[serde(default)]
  pub tags: Vec<String>,
  #[serde(default)]
  pub useFloatIP: bool,
  #[serde(default)]
  pub userData: Option<String>,
  pub tsksrvcs: Vec<TskSrvc>,
}
impl Default for ServerInstance {
  fn default() -> Self {
    let net: NetworkInterface = NetworkInterface::default();
    let net_basic: NetworkInterface = NetworkInterface::manual_basic();
    let storage = StorageDevice::manual_storage_device_ubuntu20(25);
    Self {
        cluster: Cluster::default(),
        hostname: String::from(""),
        title: String::from(""),
        loginUser: LoginUser::default(),
        metadata: true,
        plan: ServerPlan::manual,
        zone: Zone::local,
        arch: String::from("linux_amd64"),
        sshAccess: SSHAccess::default(),
        networks: vec![ net, net_basic ],
        storageDevices: vec![ storage ],
        timeZone: String::from("UTC"),
        tags: Vec::new(),
        useFloatIP: false,
        userData: Some(String::from("")),
        tsksrvcs: Vec::new(),
    }  
  }
}

#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize)]
pub struct SrvrCfg { 
  #[serde(default)]
  pub role: String,
  #[serde(default)]
  pub sshUser: String,
  pub sshKeys: Vec<String>,
  #[serde(default)]
  pub sshKeyPath: String,
  pub storageDeviceSize: u16,
  #[serde(default)]
  pub tags: Vec<String>,
  #[serde(default)]
  pub useFloatIP: bool
}
impl Default for SrvrCfg {
  fn default() -> Self {
    Self {
      role: String::from(""),
      sshUser: String::from(""),
      sshKeys: Vec::new(),
      sshKeyPath: String::from("/root/.ssh/id_rsa"),
      storageDeviceSize: 25,
      tags: Vec::new(),
      useFloatIP: false
    }
  }
}

#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct ResourcesConfig {
  pub floatIP: Option<String>,
  #[serde(default)]
  pub hostPrefix: String,
  #[serde(default)]
  pub mainName: String,
	pub group: String,
  pub group_path: String,
  pub provider: Option<String>,
  pub domainName: Option<String>,
  pub cntrllrs: Vec<Cntrllr>,
  #[serde(default)]
  pub srvrcfg: SrvrCfg,
  #[serde(default)]
  pub servers: Vec<ServerInstance>,
}

#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct ConfigResources {
  pub floatIP: Option<String>,
  #[serde(default)]
  pub hostPrefix: String,
  #[serde(default)]
  pub mainName: String,
	pub group: String,
  pub group_path: String,
  pub provider: Option<String>,
  pub domainName: Option<String>,
  pub cntrllrs: Vec<Cntrllr>,
  #[serde(default)]
  pub servers: Vec<ServerInstance>,
}

// Implementations & settings  

impl StorageDevice {
  pub fn manual_storage_device_ubuntu20 (size: u16) -> Self {
    Self {
      action: String::from("clone"),
      storage: String::from("01000000-0000-4000-8000-000030200200"),
      size: size,
      finalSize: size,
      partSizes: Some(String::from("")),
      makefs: Some(String::from("")),
      title: String::from("Ubuntu Server 20.04 LTS (Focal Fossa)"),
      tier: String::from("maxiopsa"),
      source: String::from(""),
    }
  }
}

impl NetworkInterface {
  pub fn manual_basic () -> Self {
    let mut basic =  NetworkInterface::default();
    basic.access = String::from("utility");
    basic
  }
  pub fn manual_sdn () -> Self {
    Self {
      access: String::from("private"),
      family: String::from("IPv4"),
      network: Some(String::from("03fa354b-3896-4d04-9f54-d8c7b8da654e")),
      ipaddress: Some(String::from("")),
    }
  }
}

impl ServerInstance {
  pub fn server (plan: ServerPlan, zone: Zone, size: u16, arch: &str) -> Self {
    let net: NetworkInterface = NetworkInterface::default();
    let net_basic: NetworkInterface = NetworkInterface::manual_basic();
    let storage = StorageDevice::manual_storage_device_ubuntu20(size);
    Self {
      cluster: Cluster::default(),
      hostname: String::from(""),
      title: String::from(""),
      loginUser: LoginUser::default(),
      metadata: true,
      plan,
      zone,
      arch:  String::from(arch),
      sshAccess: SSHAccess::default(),
      networks: vec![ net, net_basic ],
      storageDevices: vec![ storage ],
      timeZone: String::from("UTC"),
      tags: Vec::new(),
      useFloatIP: false,
      userData: Some(String::from("")),
      tsksrvcs: Vec::new(),
    }  
  }
}
